from abc import abstractmethod
from typing import Optional
from unittest.mock import MagicMock

from tmtccmd.com import ComInterface
from tmtccmd.config.com import ComCfgBase
from tmtccmd.config.tmtc import TmtcDefinitionWrapper
from tmtccmd.core.ccsds_backend import CcsdsTmtcBackend
from tmtccmd.config import HookBase
from tmtccmd.util.obj_id import ObjectIdDictT


def create_hook_mock() -> HookBase:
    """Create simple minimal hook mock using the MagicMock facilities by unittest
    :return:
    """
    tmtc_hook_base = MagicMock(spec=HookBase)
    return tmtc_hook_base


def create_hook_mock_with_srv_handlers() -> HookBase:
    tmtc_hook_base = create_hook_mock()
    tmtc_hook_base.handle_service_8_telemetry = MagicMock(return_value=(["Test"], [0]))
    # Valid returnvalue for now
    srv_3_return_tuple = (["Test"], [0], bytearray(0b10000000), 1)
    tmtc_hook_base.handle_service_3_housekeeping = MagicMock(
        return_value=srv_3_return_tuple
    )
    return tmtc_hook_base


class TestHookObj(HookBase):
    service_8_handler_called = False
    service_5_handler_called = False
    service_3_handler_called = False

    def __init__(self):
        super().__init__()
        self.get_obj_id_called = False
        self.add_globals_pre_args_parsing_called = False
        self.add_globals_post_args_parsing_called = False
        self.assign_communication_interface_called = False

    @abstractmethod
    def get_object_ids(self) -> ObjectIdDictT:
        """The user can specify an object ID dictionary here mapping object ID bytearrays to a
        list. This list could contain containing the string representation or additional
        information about that object ID.
        """
        return super().get_object_ids()

    @abstractmethod
    def assign_communication_interface(self, com_if_key: str) -> Optional[ComInterface]:
        """Assign the communication interface used by the TMTC commander to send and receive
        TMTC with.

        :param com_if_key:      String key of the communication interface to be created.
        """
        from tmtccmd.config.com import create_com_interface_default

        com_cfg = ComCfgBase(com_if_key, self.cfg_path)
        return create_com_interface_default(com_cfg)

    @abstractmethod
    def get_tmtc_definitions(self) -> TmtcDefinitionWrapper:
        """This is a dicitonary mapping services represented by strings to an operation code
        dictionary.

        :return:
        """
        from tmtccmd.config.globals import get_default_tmtc_defs

        return get_default_tmtc_defs()

    @abstractmethod
    def perform_mode_operation(self, tmtc_backend: CcsdsTmtcBackend, mode: int):
        """Perform custom mode operations
        :param tmtc_backend:
        :param mode:
        :return:
        """
        pass



import argparse
from unittest import TestCase

from tests.hook_obj_mock import create_hook_mock
from tmtccmd import CoreModeList, CoreModeConverter
from tmtccmd.config.args import args_to_params_tmtc, SetupParams, DefaultProcedureParams


class TestArgs(TestCase):
    def setUp(self) -> None:
        self.pargs = argparse.Namespace()
        self.params = SetupParams()
        self.hook_mock = create_hook_mock()

    def base_cli_set(self):
        self.pargs.gui = None
        self.pargs.mode = None
        self.pargs.delay = None
        self.pargs.com_if = "dummy"
        self.pargs.listener = True
        self.pargs.prompt_proc = False
        self.pargs.service = None
        self.pargs.op_code = None

    def simple_pargs_cli_set(self):
        self.base_cli_set()
        self.pargs.listener = False
        self.pargs.service = "17"
        self.pargs.op_code = "ping"

    def auto_listener_cli_set(self):
        self.base_cli_set()
        self.pargs.listener = True

    def test_basic(self):
        # For some reason, those fields need to be reset manually
        self.params.backend_params.mode = ""
        self.params.backend_params.com_if_id = ""
        self.simple_pargs_cli_set()
        self.assertEqual(self.params.tc_params.delay, 0)
        self.assertEqual(self.params.backend_params.mode, "")
        self.assertEqual(self.params.backend_params.com_if_id, "")
        def_params = DefaultProcedureParams()
        args_to_params_tmtc(
            pargs=self.pargs,
            params=self.params,
            hook_obj=self.hook_mock,
            use_prompts=False,
            def_tmtc_params=def_params,
            assign_com_if=False,
        )
        # Set to default value
        self.assertEqual(self.params.tc_params.delay, 4.0)
        # Unset
        self.assertEqual(self.params.tc_params.apid, 0)
        self.assertEqual(self.params.app_params.use_gui, False)
        self.assertEqual(self.params.app_params.use_ansi_colors, True)
        self.assertEqual(def_params.service, "17")
        self.assertEqual(def_params.op_code, "ping")
        self.assertEqual(
            self.params.backend_params.mode,
            CoreModeConverter.get_str(CoreModeList.ONE_QUEUE_MODE),
        )
        self.assertEqual(self.params.backend_params.listener, False)
        self.assertEqual(self.params.backend_params.com_if_id, "dummy")

    def test_delay_set(self):
        self.simple_pargs_cli_set()
        self.pargs.delay = 2.0
        def_params = DefaultProcedureParams()
        args_to_params_tmtc(
            pargs=self.pargs,
            params=self.params,
            hook_obj=self.hook_mock,
            use_prompts=False,
            def_tmtc_params=def_params,
            assign_com_if=False,
        )
        self.assertEqual(def_params.service, "17")
        self.assertEqual(def_params.op_code, "ping")
        self.assertEqual(self.params.tc_params.delay, 2.0)

    def test_auto_listener_mode(self):
        self.auto_listener_cli_set()
        def_params = DefaultProcedureParams()
        args_to_params_tmtc(
            pargs=self.pargs,
            params=self.params,
            hook_obj=self.hook_mock,
            use_prompts=False,
            def_tmtc_params=def_params,
            assign_com_if=False,
        )
        self.assertEqual(self.params.backend_params.listener, True)
        self.assertIsNone(def_params.service)
        self.assertIsNone(def_params.op_code)
        self.assertEqual(
            self.params.backend_params.mode,
            CoreModeConverter.get_str(CoreModeList.LISTENER_MODE),
        )
        self.assertEqual(
            self.params.mode, CoreModeConverter.get_str(CoreModeList.LISTENER_MODE)
        )
        self.params = SetupParams()



from dataclasses import dataclass
from pathlib import Path

from spacepackets.cfdp import Direction, ChecksumType
from spacepackets.util import UnsignedByteField


class NoRemoteEntityCfgFound(Exception):
    def __init__(self, entity_id: UnsignedByteField, *args, **kwargs):
        super().__init__(args, kwargs)
        self.remote_entity_id = entity_id

    def __str__(self):
        return f"No remote entity found for entity ID {self.remote_entity_id}"


class SourceFileDoesNotExist(Exception):
    def __init__(self, file: Path, *args, **kwargs):
        super().__init__(args, kwargs)
        self.file = file

    def __str__(self):
        return f"Source file {self.file} does not exist"


class ChecksumNotImplemented(Exception):
    def __init__(self, checksum_type: ChecksumType, *args, **kwargs):
        super().__init__(args, kwargs)
        self.checksum_type = checksum_type

    def __str__(self):
        return f"{self.checksum_type} not implemented"


class PacketSendNotConfirmed(Exception):
    pass


class InvalidPduDirection(Exception):
    def __init__(self, expected_dir: Direction, found_dir: Direction, *args, **kwargs):
        super().__init__(args, kwargs)
        self.expected_dir = expected_dir
        self.found_dir = found_dir

    def __str__(self):
        return f"Expected direction {self.expected_dir}, got {self.found_dir}"


class InvalidSourceId(Exception):
    """Invalid source entity ID. This is not necessarily the sender of a packet but actually the
    entity that started a transaction, or the entity which is transferring a file"""

    def __init__(
        self,
        expected_src_id: UnsignedByteField,
        found_src_id: UnsignedByteField,
        *args,
        **kwargs,
    ):
        super().__init__(args, kwargs)
        self.expected_src_id = expected_src_id
        self.found_src_id = found_src_id

    def __str__(self):
        return f"Expected source {self.expected_src_id}, got {self.found_src_id}"


class InvalidDestinationId(Exception):
    """Invalid destination entity ID. This is not necessarily the receiver of a packet but actually
    the recipient of a file, or the entity receiving file data and metadata PDUs"""

    def __init__(
        self,
        expected_dest_id: UnsignedByteField,
        found_dest_id: UnsignedByteField,
        *args,
        **kwargs,
    ):
        super().__init__(args, kwargs)
        self.expected_dest_id = expected_dest_id
        self.found_dest_id = found_dest_id

    def __str__(self):
        return f"Expected destination {self.expected_dest_id}, got {self.found_dest_id}"


class BusyError(Exception):
    pass


@dataclass
class FileParamsBase:
    progress: int
    segment_len: int
    crc32: bytes
    no_file_data: bool
    file_size: int

    @classmethod
    def empty(cls):
        return cls(
            progress=0,
            segment_len=0,
            crc32=bytes(),
            file_size=0,
            no_file_data=False,
        )

    def reset(self):
        self.progress = 0
        self.segment_len = 0
        self.crc32 = bytes()
        self.file_size = 0



"""Contains definitions and functions related to PUS Service 5 Telecommands.
"""
import deprecation

from tmtccmd import __version__
from spacepackets.ecss import PusTelecommand
from spacepackets.ecss.pus_5_event import Subservice


@deprecation.deprecated(
    deprecated_in="v4.0.0a0",
    current_version=__version__,
    details="use create... API instead",
)
def pack_enable_event_reporting_command(
    apid: int = 0, seq_count: int = 0
) -> PusTelecommand:
    return create_disable_event_reporting_command(apid, seq_count)


def create_enable_event_reporting_command(
    apid: int = 0, seq_count: int = 0
) -> PusTelecommand:
    return PusTelecommand(
        service=5,
        subservice=Subservice.TC_ENABLE_EVENT_REPORTING,
        seq_count=seq_count,
        apid=apid,
    )


@deprecation.deprecated(
    deprecated_in="v4.0.0a0",
    current_version=__version__,
    details="use create... API instead",
)
def pack_disable_event_reporting_command(
    apid: int = 0, seq_count: int = 0
) -> PusTelecommand:
    return create_disable_event_reporting_command(apid, seq_count)


def create_disable_event_reporting_command(
    apid: int = 0, seq_count: int = 0
) -> PusTelecommand:
    return PusTelecommand(
        service=5,
        subservice=Subservice.TC_DISABLE_EVENT_REPORTING,
        seq_count=seq_count,
        apid=apid,
    )



from typing import Union, List, Optional, Dict, Tuple

ServiceNameT = str
ServiceInfoT = str
OpCodeNameT = Union[str, List[str]]
OpCodeInfoT = str


class OpCodeOptionBase:
    def __init__(self):
        pass


OpCodeDict = Dict[str, Tuple[OpCodeInfoT, OpCodeOptionBase]]


class OpCodeEntry:
    def __init__(self):
        self._op_code_dict_num_keys: OpCodeDict = dict()
        self._op_code_dict_str_keys: OpCodeDict = dict()

    def add(
        self,
        keys: OpCodeNameT,
        info: str,
        options: OpCodeOptionBase = OpCodeOptionBase(),
    ):
        if isinstance(keys, str):
            keys = [keys]
        for key in keys:
            if key.isdigit():
                self._op_code_dict_num_keys.update({key: (info, options)})
            else:
                self._op_code_dict_str_keys.update({key: (info, options)})

    def sort_num_key_dict(self):
        self._op_code_dict_num_keys = {
            int(k): v for k, v in self._op_code_dict_num_keys.items()
        }

    def sort_text_key_dict(self):
        self._op_code_dict_str_keys = {
            k: v for k, v in sorted(self._op_code_dict_str_keys.items())
        }

    def info(self, op_code: str) -> Optional[str]:
        if op_code.isdigit():
            entry_tuple = self._op_code_dict_num_keys.get(op_code)
            if entry_tuple is not None:
                return entry_tuple[0]

        entry_tuple = self._op_code_dict_str_keys.get(op_code)
        if entry_tuple is not None:
            return entry_tuple[0]
        return None

    def __str__(self):
        return (
            f"Op codes with numeric keys: {self._op_code_dict_num_keys!r}), "
            f"op codes with text keys: {self._op_code_dict_str_keys!r}"
        )

    @property
    def op_code_dict_num_keys(self):
        return self._op_code_dict_num_keys

    @property
    def op_code_dict_str_keys(self):
        return self._op_code_dict_str_keys


# It is possible to specify a service without any op codes
ServiceDictValueT = Optional[Tuple[ServiceInfoT, OpCodeEntry]]
ServiceOpCodeDictT = Dict[ServiceNameT, ServiceDictValueT]


class TmtcDefinitionWrapper:
    def __init__(self, init_defs: Optional[ServiceOpCodeDictT] = None):
        if init_defs is None:
            self.defs: ServiceOpCodeDictT = dict()
        else:
            self.defs = init_defs

    def __repr__(self):
        return f"{self.__class__.__name__}(init_defs={self.defs!r}"

    def add_service(
        self,
        name: str,
        info: str,
        op_code_entry: OpCodeEntry,
    ):
        self.defs.update({name: (info, op_code_entry)})

    def op_code_entry(self, service_name: str) -> Optional[OpCodeEntry]:
        srv_entry = self.defs.get(service_name)
        if srv_entry is not None:
            return srv_entry[1]
        return None

    def sort(self):
        self.defs = {key: self.defs[key] for key in sorted(self.defs.keys())}


REGISTER_CBS = set()


def tmtc_definitions_provider(adder_func):
    """Function decorator which registers the decorated function to be a TMTC definition provider.
    The :py:func:`execute_tmtc_def_functions` function can be used to call all functions.

    It is expected that the decorated function takes the py:class:`TmTcDefWrapper` as the first
    argument. The user can pass any additional arguments as positional and keyword
    arguments.
    """
    global REGISTER_CBS

    def call_explicitely(defs: TmtcDefinitionWrapper, *args, **kwargs):
        if REGISTER_CBS and adder_func in REGISTER_CBS:
            REGISTER_CBS.remove(adder_func)
        print(f"{adder_func.__name__} called explicitely, not necessary")
        return adder_func(defs, args, kwargs)

    REGISTER_CBS.add(adder_func)
    return call_explicitely


def call_all_definitions_providers(defs: TmtcDefinitionWrapper, *args, **kwargs):
    global REGISTER_CBS
    if REGISTER_CBS:
        for register_cb in REGISTER_CBS.copy():
            if args and kwargs:
                register_cb(defs, args, kwargs)
            if args and not kwargs:
                register_cb(defs, args)
            elif not args and not kwargs:
                register_cb(defs)
            REGISTER_CBS.remove(register_cb)



import dataclasses
import os
import random
import struct
import sys
from crcmod.predefined import mkPredefinedCrcFun
import tempfile
from typing import cast, Optional
from pathlib import Path
from unittest import TestCase
from unittest.mock import MagicMock

from spacepackets.cfdp import (
    ChecksumType,
    PduConfig,
    TransmissionMode,
    NULL_CHECKSUM_U32,
    ConditionCode,
)
from spacepackets.cfdp.pdu import (
    MetadataPdu,
    MetadataParams,
    EofPdu,
    FileDataPdu,
    FileDeliveryStatus,
)
from spacepackets.cfdp.pdu.file_data import FileDataParams
from spacepackets.util import ByteFieldU16, ByteFieldU8
from tmtccmd.cfdp import (
    IndicationCfg,
    LocalEntityCfg,
    RemoteEntityCfgTable,
    RemoteEntityCfg,
)
from tmtccmd.cfdp.defs import CfdpStates, TransactionId
from tmtccmd.cfdp.handler.dest import (
    DestHandler,
    TransactionStep,
    FsmResult,
)
from tmtccmd.cfdp.user import TransactionFinishedParams, FileSegmentRecvdParams

from .cfdp_fault_handler_mock import FaultHandler
from .cfdp_user_mock import CfdpUser


@dataclasses.dataclass
class FileInfo:
    rand_data: bytes
    file_size: int
    crc32: bytes


class TestCfdpDestHandler(TestCase):
    def setUp(self) -> None:
        self.indication_cfg = IndicationCfg(True, True, True, True, True, True)
        self.fault_handler = FaultHandler()
        self.entity_id = ByteFieldU16(2)
        self.local_cfg = LocalEntityCfg(
            self.entity_id, self.indication_cfg, self.fault_handler
        )
        self.src_entity_id = ByteFieldU16(1)
        self.src_pdu_conf = PduConfig(
            source_entity_id=self.src_entity_id,
            dest_entity_id=self.entity_id,
            transaction_seq_num=ByteFieldU8(1),
            trans_mode=TransmissionMode.UNACKNOWLEDGED,
        )
        self.transaction_id = TransactionId(self.src_entity_id, ByteFieldU8(1))
        self.closure_requested = False
        self.cfdp_user = CfdpUser()
        self.file_segment_len = 128
        self.cfdp_user.eof_recv_indication = MagicMock()
        self.cfdp_user.file_segment_recv_indication = MagicMock()
        self.cfdp_user.transaction_finished_indication = MagicMock()
        self.src_file_path = Path(f"{tempfile.gettempdir()}/hello.txt")
        if self.src_file_path.exists():
            os.remove(self.src_file_path)
        self.dest_file_path = Path(f"{tempfile.gettempdir()}/hello_dest.txt")
        if self.dest_file_path.exists():
            os.remove(self.dest_file_path)
        self.remote_cfg_table = RemoteEntityCfgTable()
        self.remote_cfg = RemoteEntityCfg(
            entity_id=self.src_entity_id,
            check_limit=None,
            crc_type=ChecksumType.CRC_32,
            closure_requested=False,
            crc_on_transmission=False,
            default_transmission_mode=TransmissionMode.UNACKNOWLEDGED,
            max_file_segment_len=self.file_segment_len,
        )
        self.remote_cfg_table.add_config(self.remote_cfg)
        self.dest_handler = DestHandler(
            self.local_cfg, self.cfdp_user, self.remote_cfg_table
        )

    def test_empty_file_reception(self):
        metadata_params = MetadataParams(
            checksum_type=ChecksumType.NULL_CHECKSUM,
            closure_requested=False,
            source_file_name=self.src_file_path.as_posix(),
            dest_file_name=self.dest_file_path.as_posix(),
            file_size=0,
        )
        file_transfer_init = MetadataPdu(
            params=metadata_params, pdu_conf=self.src_pdu_conf
        )
        self._state_checker(None, CfdpStates.IDLE, TransactionStep.IDLE)
        self.dest_handler.pass_packet(file_transfer_init)
        fsm_res = self.dest_handler.state_machine()
        self.assertFalse(fsm_res.states.packet_ready)
        self.assertEqual(
            self.dest_handler.states.transaction, TransactionStep.RECEIVING_FILE_DATA
        )
        eof_pdu = EofPdu(
            file_size=0, file_checksum=NULL_CHECKSUM_U32, pdu_conf=self.src_pdu_conf
        )
        self.dest_handler.pass_packet(eof_pdu)
        fsm_res = self.dest_handler.state_machine()
        self._state_checker(fsm_res, CfdpStates.IDLE, TransactionStep.IDLE)
        self._check_eof_recv_indication(fsm_res)
        self._check_finished_recv_indication_success(fsm_res)
        self.assertTrue(self.dest_file_path.exists())
        self.assertEqual(self.dest_file_path.stat().st_size, 0)

    def test_small_file_reception(self):
        data = "Hello World\n".encode()
        with open(self.src_file_path, "wb") as of:
            of.write(data)
        crc32_func = mkPredefinedCrcFun("crc32")
        crc32 = struct.pack("!I", crc32_func(data))
        file_size = self.src_file_path.stat().st_size
        self._source_simulator_transfer_init_with_metadata(
            checksum=ChecksumType.CRC_32,
            file_size=file_size,
            file_path=self.src_file_path.as_posix(),
        )
        with open(self.src_file_path, "rb") as rf:
            read_data = rf.read()
        fd_params = FileDataParams(file_data=read_data, offset=0)
        file_data_pdu = FileDataPdu(params=fd_params, pdu_conf=self.src_pdu_conf)
        self.dest_handler.pass_packet(file_data_pdu)
        fsm_res = self.dest_handler.state_machine()
        self._state_checker(
            fsm_res, CfdpStates.BUSY_CLASS_1_NACKED, TransactionStep.RECEIVING_FILE_DATA
        )
        eof_pdu = EofPdu(
            file_size=file_size,
            file_checksum=crc32,
            pdu_conf=self.src_pdu_conf,
        )
        self.dest_handler.pass_packet(eof_pdu)
        fsm_res = self.dest_handler.state_machine()
        self._state_checker(fsm_res, CfdpStates.IDLE, TransactionStep.IDLE)
        self._check_eof_recv_indication(fsm_res)
        self._check_finished_recv_indication_success(fsm_res)

    def test_larger_file_reception(self):
        # This tests generates two file data PDUs, but the second one does not have a
        # full segment length
        file_info = self.random_data_two_file_segments()
        self._state_checker(None, CfdpStates.IDLE, TransactionStep.IDLE)
        self._source_simulator_transfer_init_with_metadata(
            checksum=ChecksumType.CRC_32,
            file_size=file_info.file_size,
            file_path=self.src_file_path.as_posix(),
        )
        fsm_res = self.pass_file_segment(
            file_info.rand_data[0 : self.file_segment_len], 0
        )
        self._state_checker(
            fsm_res, CfdpStates.BUSY_CLASS_1_NACKED, TransactionStep.RECEIVING_FILE_DATA
        )
        self.cfdp_user.file_segment_recv_indication.assert_called()
        self.assertEqual(self.cfdp_user.file_segment_recv_indication.call_count, 1)
        seg_recv_params = cast(
            FileSegmentRecvdParams,
            self.cfdp_user.file_segment_recv_indication.call_args.args[0],
        )
        self.assertEqual(seg_recv_params.transaction_id, self.transaction_id)
        fd_params = FileDataParams(
            file_data=file_info.rand_data[self.file_segment_len :],
            offset=self.file_segment_len,
        )
        file_data_pdu = FileDataPdu(params=fd_params, pdu_conf=self.src_pdu_conf)
        self.dest_handler.pass_packet(file_data_pdu)
        fsm_res = self.dest_handler.state_machine()
        self._state_checker(
            fsm_res, CfdpStates.BUSY_CLASS_1_NACKED, TransactionStep.RECEIVING_FILE_DATA
        )
        eof_pdu = EofPdu(
            file_size=file_info.file_size,
            file_checksum=file_info.crc32,
            pdu_conf=self.src_pdu_conf,
        )
        self.dest_handler.pass_packet(eof_pdu)
        fsm_res = self.dest_handler.state_machine()
        self._state_checker(fsm_res, CfdpStates.IDLE, TransactionStep.IDLE)
        self._check_eof_recv_indication(fsm_res)
        self._check_finished_recv_indication_success(fsm_res)

    def random_data_two_file_segments(self):
        if sys.version_info >= (3, 9):
            rand_data = random.randbytes(round(self.file_segment_len * 1.3))
        else:
            rand_data = os.urandom(round(self.file_segment_len * 1.3))
        file_size = len(rand_data)
        crc32_func = mkPredefinedCrcFun("crc32")
        crc32 = struct.pack("!I", crc32_func(rand_data))
        return FileInfo(file_size=file_size, crc32=crc32, rand_data=rand_data)

    def test_file_is_overwritten(self):
        with open(self.dest_file_path, "w") as of:
            of.write("This file will be truncated")
        self.test_small_file_reception()

    def test_file_data_pdu_before_metadata_is_discarded(self):
        file_info = self.random_data_two_file_segments()
        # Pass file data PDU first. Will be discarded
        fsm_res = self.pass_file_segment(
            file_info.rand_data[0 : self.file_segment_len], 0
        )
        self._state_checker(fsm_res, CfdpStates.IDLE, TransactionStep.IDLE)
        self._source_simulator_transfer_init_with_metadata(
            checksum=ChecksumType.CRC_32,
            file_size=file_info.file_size,
            file_path=self.src_file_path.as_posix(),
        )
        fsm_res = self.pass_file_segment(
            segment=file_info.rand_data[self.file_segment_len :],
            offset=self.file_segment_len,
        )
        self._state_checker(
            fsm_res, CfdpStates.BUSY_CLASS_1_NACKED, TransactionStep.RECEIVING_FILE_DATA
        )
        eof_pdu = EofPdu(
            file_size=file_info.file_size,
            file_checksum=file_info.crc32,
            pdu_conf=self.src_pdu_conf,
        )
        self.dest_handler.pass_packet(eof_pdu)
        fsm_res = self.dest_handler.state_machine()
        self.cfdp_user.transaction_finished_indication.assert_called_once()
        finished_args = cast(
            TransactionFinishedParams,
            self.cfdp_user.transaction_finished_indication.call_args.args[0],
        )
        # At least one segment was stored
        self.assertEqual(finished_args.file_status, FileDeliveryStatus.FILE_RETAINED)
        self.assertEqual(
            finished_args.condition_code, ConditionCode.FILE_CHECKSUM_FAILURE
        )
        self._state_checker(fsm_res, CfdpStates.IDLE, TransactionStep.IDLE)

    def test_permission_error(self):
        with open(self.src_file_path, "w") as of:
            of.write("Hello World\n")
        self.src_file_path.chmod(0o444)
        # TODO: This will cause permission errors, but the error handling for this has not been
        #       implemented properly
        """
        file_size = src_file.stat().st_size
        self._source_simulator_transfer_init_with_metadata(
            checksum=ChecksumTypes.CRC_32,
            file_size=file_size,
            file_path=src_file.as_posix(),
        )
        with open(src_file, "rb") as rf:
            read_data = rf.read()
        fd_params = FileDataParams(file_data=read_data, offset=0)
        file_data_pdu = FileDataPdu(params=fd_params, pdu_conf=self.src_pdu_conf)
        self.dest_handler.pass_packet(file_data_pdu)
        fsm_res = self.dest_handler.state_machine()
        self._state_checker(
            fsm_res, CfdpStates.BUSY_CLASS_1_NACKED, TransactionStep.RECEIVING_FILE_DATA
        )
        """
        self.src_file_path.chmod(0o777)

    def _check_eof_recv_indication(self, fsm_res: FsmResult):
        self.cfdp_user.eof_recv_indication.assert_called_once()
        self.assertEqual(
            self.cfdp_user.eof_recv_indication.call_args.args[0], self.transaction_id
        )
        self.assertEqual(fsm_res.states.transaction_id, self.transaction_id)

    def _check_finished_recv_indication_success(self, fsm_res: FsmResult):
        finished_params = cast(
            TransactionFinishedParams,
            self.cfdp_user.transaction_finished_indication.call_args.args[0],
        )
        self.assertEqual(finished_params.transaction_id, self.transaction_id)
        self.assertEqual(fsm_res.states.transaction_id, self.transaction_id)
        self.assertEqual(finished_params.condition_code, ConditionCode.NO_ERROR)

    def pass_file_segment(self, segment: bytes, offset) -> FsmResult:
        fd_params = FileDataParams(file_data=segment, offset=offset)
        file_data_pdu = FileDataPdu(params=fd_params, pdu_conf=self.src_pdu_conf)
        self.dest_handler.pass_packet(file_data_pdu)
        return self.dest_handler.state_machine()

    def _state_checker(
        self,
        fsm_res: Optional[FsmResult],
        expected_state: CfdpStates,
        expected_transaction: TransactionStep,
    ):
        if fsm_res is not None:
            self.assertEqual(fsm_res.states.state, expected_state)
            self.assertEqual(fsm_res.states.transaction, expected_transaction)
        self.assertEqual(self.dest_handler.states.state, expected_state)
        self.assertEqual(self.dest_handler.states.transaction, expected_transaction)

    def _source_simulator_transfer_init_with_metadata(
        self, checksum: ChecksumType, file_path: str, file_size: int
    ):
        """A file transfer on the receiving side is always initiated by sending a metadata PDU.
        This function simulates a CFDP source entity which initiates a file transfer by sending
        this PDU.
        """
        metadata_params = MetadataParams(
            checksum_type=checksum,
            closure_requested=self.closure_requested,
            source_file_name=file_path,
            dest_file_name=self.dest_file_path.as_posix(),
            file_size=file_size,
        )
        file_transfer_init = MetadataPdu(
            params=metadata_params, pdu_conf=self.src_pdu_conf
        )
        self._state_checker(None, CfdpStates.IDLE, TransactionStep.IDLE)
        self.dest_handler.pass_packet(file_transfer_init)

    def tearDown(self) -> None:
        # self.dest_handler.finish()
        if self.dest_file_path.exists():
            os.remove(self.dest_file_path)
        if self.src_file_path.exists():
            os.remove(self.src_file_path)



import enum
from typing import Any, cast, Type, Optional

from tmtccmd.cfdp import CfdpRequestWrapper


class TcProcedureType(enum.Enum):
    DEFAULT = 0
    CFDP = 1
    CUSTOM = 2


class TcProcedureBase:
    def __init__(self, ptype: TcProcedureType):
        self.ptype = ptype


class CustomProcedureInfo(TcProcedureBase):
    def __init__(self, info: any):
        super().__init__(TcProcedureType.CUSTOM)
        self.info = info

    def __repr__(self):
        return f"{self.__class__.__name__}(info={self.info!r}"


class DefaultProcedureInfo(TcProcedureBase):
    """Generic abstraction for procedures. A procedure can be a single command or a sequence
    of commands. Generally, one procedure is mapped to a specific TC queue which is packed
    during run-time"""

    def __init__(self, service: str, op_code: str):
        super().__init__(TcProcedureType.DEFAULT)
        self.service = service
        self.op_code = op_code

    @classmethod
    def empty(cls):
        return cls("", "")

    def __repr__(self):
        return f"CmdInfo(service={self.service!r}, op_code={self.op_code!r})"


class CfdpProcedureInfo(TcProcedureBase):
    def __init__(self):
        super().__init__(TcProcedureType.CFDP)
        self.request_wrapper = CfdpRequestWrapper(None)

    @property
    def cfdp_request_type(self):
        return self.request_wrapper.request


class ProcedureWrapper:
    """Procedure helper class. It wraps the concrete procedure object but allows easily casting
    it to concrete types supported by the framework."""

    def __init__(self, base: Optional[TcProcedureBase]):
        self.base = base

    def __repr__(self):
        return f"{self.__class__.__name__}(base={self.base!r})"

    @property
    def proc_type(self):
        return self.base.ptype

    def __cast_internally(
        self,
        obj_type: Type[TcProcedureBase],
        obj: TcProcedureBase,
        expected_type: TcProcedureType,
    ) -> Any:
        if obj.ptype != expected_type:
            raise TypeError(f"Invalid object {obj} for type {self.base.ptype}")
        return cast(obj_type, obj)

    def to_def_procedure(self) -> DefaultProcedureInfo:
        return self.__cast_internally(
            DefaultProcedureInfo, self.base, TcProcedureType.DEFAULT
        )

    def to_cfdp_procedure(self) -> CfdpProcedureInfo:
        return self.__cast_internally(
            CfdpProcedureInfo, self.base, TcProcedureType.CFDP
        )

    def to_custom_procedure(self) -> CustomProcedureInfo:
        return self.__cast_internally(
            CustomProcedureInfo, self.base, TcProcedureType.CUSTOM
        )



import enum
import logging
from dataclasses import dataclass
from typing import Optional, Dict, List

from spacepackets.cfdp import (
    TransmissionMode,
    NULL_CHECKSUM_U32,
    ConditionCode,
    Direction,
    PduConfig,
    ChecksumType,
    FaultHandlerCode,
)
from spacepackets.cfdp.pdu import (
    PduHolder,
    FileDeliveryStatus,
    DeliveryCode,
    EofPdu,
    FileDataPdu,
    MetadataPdu,
    MetadataParams,
    DirectiveType,
    AbstractFileDirectiveBase,
)
from spacepackets.cfdp.pdu.file_data import FileDataParams
from spacepackets.util import UnsignedByteField, ByteFieldGenerator
from tmtccmd.cfdp import (
    LocalEntityCfg,
    CfdpUserBase,
    TransactionId,
    RemoteEntityCfg,
)
from tmtccmd.cfdp.defs import CfdpRequestType, CfdpStates
from tmtccmd.cfdp.filestore import VirtualFilestore
from tmtccmd.cfdp.handler.crc import Crc32Helper
from tmtccmd.cfdp.handler.defs import (
    FileParamsBase,
    PacketSendNotConfirmed,
    SourceFileDoesNotExist,
    InvalidPduDirection,
    InvalidSourceId,
    InvalidDestinationId,
)
from tmtccmd.cfdp.mib import EntityType
from tmtccmd.cfdp.request import CfdpRequestWrapper, PutRequest
from tmtccmd.cfdp.user import TransactionFinishedParams
from tmtccmd.util import ProvidesSeqCount
from tmtccmd.util.countdown import Countdown

_LOGGER = logging.getLogger(__name__)


class TransactionStep(enum.Enum):
    IDLE = 0
    TRANSACTION_START = 1
    CRC_PROCEDURE = 2
    # The following three are used for the Copy File Procedure
    SENDING_METADATA = 3
    SENDING_FILE_DATA = 4
    SENDING_EOF = 5
    WAIT_FOR_ACK = 6
    WAIT_FOR_FINISH = 7
    NOTICE_OF_COMPLETION = 8


@dataclass
class SourceStateWrapper:
    state: CfdpStates = CfdpStates.IDLE
    step: TransactionStep = TransactionStep.IDLE
    packet_ready: bool = False


class TransferFieldWrapper:
    def __init__(self, local_entity_id: UnsignedByteField, vfs: VirtualFilestore):
        self.crc_helper = Crc32Helper(ChecksumType.NULL_CHECKSUM, vfs)
        self.transaction: Optional[TransactionId] = None
        self.check_limit: Optional[Countdown] = None
        self.fp = FileParamsBase.empty()
        self.remote_cfg: Optional[RemoteEntityCfg] = None
        self.closure_requested: bool = False
        self.pdu_conf = PduConfig.empty()
        self.pdu_conf.source_entity_id = local_entity_id

    @property
    def source_id(self):
        return self.pdu_conf.source_entity_id

    @source_id.setter
    def source_id(self, source_id: UnsignedByteField):
        self.pdu_conf.source_entity_id = source_id

    @property
    def dest_id(self):
        return self.pdu_conf.dest_entity_id

    @dest_id.setter
    def dest_id(self, dest_id: UnsignedByteField):
        self.pdu_conf.dest_entity_id = dest_id

    @property
    def transmission_mode(self) -> TransmissionMode:
        return self.pdu_conf.trans_mode

    @transmission_mode.setter
    def transmission_mode(self, trans_mode: TransmissionMode):
        self.pdu_conf.trans_mode = trans_mode

    @property
    def transaction_seq_num(self) -> UnsignedByteField:
        return self.pdu_conf.transaction_seq_num

    @transaction_seq_num.setter
    def transaction_seq_num(self, seq_num: UnsignedByteField):
        self.pdu_conf.transaction_seq_num = seq_num

    def reset(self):
        self.fp.reset()
        self.remote_cfg = None
        self.transaction = None
        self.check_limit = None
        self.closure_requested = False
        self.pdu_conf = PduConfig.empty()


class FsmResult:
    def __init__(self, pdu_holder: PduHolder, states: SourceStateWrapper):
        self.pdu_holder = pdu_holder
        self.states = states


class InvalidPduForSourceHandler(Exception):
    def __init__(self, packet: AbstractFileDirectiveBase, *args, **kwargs):
        super().__init__(args, kwargs)
        self.packet = packet

    def __str__(self):
        return f"Invalid packet {self.packet} for source handler"


class SourceHandler:
    """This is the primary CFDP source handler. It models the CFDP source entity, which is primarily
    responsible for handling put requests to send files to another CFDP destination entity.

    As such, it contains a state machine to perform all operations necessary to perform a
    source-to-destination file transfer. This class does not send the CFDP PDU packets directly
    to allow for greater flexibility. For example, a user might want to wrap the CFDP packet
    entities into a CCSDS space packet or into a special frame type. The user is responsible for
    sending the packets and confirming that they are sent successfully.

    The following core functions are the primary interface for a direct usage or for a composite
    handler with a source handler and a destination handler as member objects:

     1. :py:meth:`start_cfdp_transaction` : Can be used to start transactions, most notably to start
        and perform a Copy File procedure
     2. :py:meth:`state_machine` : This state machine generates the necessary CFDP PDUs necessary
        to perform a CFDP file transfer. The PDUs are returned in a special wrapper result type.
     3. :py:meth:`confirm_packet_sent_advance_fsm` : Confirm that the PDUs generated by the last
        state machine iteration was sent successfully and advance the state machine
     4. :py:meth:`pass_packet` : Pass reply PDUs received from a CFDP remote destination related
        to a specific transaction.

    """

    def __init__(
        self,
        cfg: LocalEntityCfg,
        seq_num_provider: ProvidesSeqCount,
        user: CfdpUserBase,
    ):
        self.states = SourceStateWrapper()
        self.pdu_holder = PduHolder(None)
        self.cfg = cfg
        self.user = user
        self.seq_num_provider = seq_num_provider
        self._params = TransferFieldWrapper(cfg.local_entity_id, self.user.vfs)
        self._current_req = CfdpRequestWrapper(None)
        self._rec_dict: Dict[DirectiveType, List[AbstractFileDirectiveBase]] = dict()

    @property
    def transaction_seq_num(self) -> UnsignedByteField:
        return self.pdu_conf.transaction_seq_num

    @property
    def pdu_conf(self) -> PduConfig:
        return self._params.pdu_conf

    @property
    def source_id(self) -> UnsignedByteField:
        return self.cfg.local_entity_id

    @source_id.setter
    def source_id(self, source_id: UnsignedByteField):
        self.cfg.local_entity_id = source_id
        self._params.source_id = source_id

    def start_cfdp_transaction(
        self, wrapper: CfdpRequestWrapper, remote_cfg: RemoteEntityCfg
    ) -> bool:
        """Start a CFDP transaction.

        :param wrapper:
        :param remote_cfg:
        :return: Whether transaction was started successfully.
        """
        if wrapper.request_type == CfdpRequestType.PUT:
            return self.put_request(wrapper.to_put_request(), remote_cfg)

    def put_request(self, request: PutRequest, remote_cfg: RemoteEntityCfg):
        if self.states.state != CfdpStates.IDLE:
            _LOGGER.debug("CFDP source handler is busy, can't process put request")
            return False
        self._current_req.base = request
        self._params.remote_cfg = remote_cfg
        self._params.dest_id = remote_cfg.entity_id
        self.states.packet_ready = False
        self._setup_transmission_mode()
        if self._params.transmission_mode == TransmissionMode.UNACKNOWLEDGED:
            _LOGGER.debug("Starting Put Request handling in NAK mode")
            self.states.state = CfdpStates.BUSY_CLASS_1_NACKED
        elif self._params.transmission_mode == TransmissionMode.ACKNOWLEDGED:
            _LOGGER.debug("Starting Put Request handling in ACK mode")
            self.states.state = CfdpStates.BUSY_CLASS_2_ACKED
        else:
            raise ValueError(
                f"Invalid transmission mode {self._params.transmission_mode} passed"
            )
        return True

    def pass_packet(self, packet: AbstractFileDirectiveBase):
        """Pass PDU file directives going towards the file sender to the CFDP source handler

        :raises InvalidPduDirection: PDU direction field wrong
        :raises InvalidPduForSourceHandler: Invalid PDU file directive type
        """
        if packet.pdu_header.direction != Direction.TOWARDS_SENDER:
            raise InvalidPduDirection(
                Direction.TOWARDS_SENDER, packet.pdu_header.direction
            )
        if packet.source_entity_id != self.source_id:
            raise InvalidSourceId(self.source_id, packet.source_entity_id)
        if packet.dest_entity_id != self._params.remote_cfg.entity_id:
            raise InvalidDestinationId(
                self._params.remote_cfg.entity_id, packet.dest_entity_id
            )
        # TODO: What about prompt and keep alive PDU?
        if packet.directive_type in [
            DirectiveType.METADATA_PDU,
            DirectiveType.EOF_PDU,
        ]:
            raise InvalidPduForSourceHandler(packet)
        # A dictionary is used to allow passing multiple received packets and store them until
        # they are processed by the state machine.
        if packet.directive_type in self._rec_dict:
            pdu_directive_list = self._rec_dict.get(packet.directive_type)
            pdu_directive_list.append(packet)
        else:
            self._rec_dict.update({packet.directive_type: [packet]})

    def __fsm_crc_procedure(self, put_req: PutRequest):
        if self._params.fp.file_size == 0:
            # Empty file, use null checksum
            self._params.fp.crc32 = NULL_CHECKSUM_U32
        else:
            self._params.fp.crc32 = self._params.crc_helper.calc_for_file(
                file=put_req.cfg.source_file,
                file_sz=self._params.fp.file_size,
                segment_len=self._params.fp.segment_len,
            )

        self.states.step = TransactionStep.SENDING_METADATA

    def __fsm_naked(  # noqa: C901  # complexity is okay here
        self,
    ) -> Optional[FsmResult]:
        put_req = self._current_req.to_put_request()
        if self.states.step == TransactionStep.IDLE:
            self.states.step = TransactionStep.TRANSACTION_START
        if self.states.step == TransactionStep.TRANSACTION_START:
            self._transaction_start(put_req)
            self.states.step = TransactionStep.CRC_PROCEDURE
        if self.states.step == TransactionStep.CRC_PROCEDURE:
            self.__fsm_crc_procedure(put_req)
        if self.states.step == TransactionStep.SENDING_METADATA:
            self._prepare_metadata_pdu(put_req)
            self.states.packet_ready = True
            return FsmResult(self.pdu_holder, self.states)
        if self.states.step == TransactionStep.SENDING_FILE_DATA:
            if self._prepare_next_file_data_pdu(put_req):
                self.states.packet_ready = True
                return FsmResult(self.pdu_holder, self.states)
            else:
                # Special case: Empty file.
                self.states.step = TransactionStep.SENDING_EOF
        if self.states.step == TransactionStep.SENDING_EOF:
            self._prepare_eof_pdu()
            self.states.packet_ready = True
            return FsmResult(self.pdu_holder, self.states)
        if self.states.step == TransactionStep.WAIT_FOR_ACK:
            self._handle_wait_for_ack()
        if self.states.step == TransactionStep.WAIT_FOR_FINISH:
            self._handle_wait_for_finish()
        if self.states.step == TransactionStep.NOTICE_OF_COMPLETION:
            self._notice_of_completion()

    def state_machine(self) -> FsmResult:
        """This is the primary state machine which performs the CFDP procedures like CRC calculation
        and PDU generation. The packets generated by this finite-state machine (FSM) need to be
        sent by the user and can be retrieved using the
        :py:class:`spacepackets.cfdp.pdu.helper.PduHolder` class contained in the
        returned :py:class:`tmtccmd.cfdp.handler.source.FsmResult`. After the packet was sent,
        the calling code has to call :py:meth:`confirm_packet_sent` and :py:meth:`advance_fsm`
        for the next state machine call do perform the next transaction step.
        There is also the helper method :py:meth:`confirm_packet_sent_advance_fsm` available
        to perform both steps.

        :raises PacketSendNotConfirmed: The FSM generated a packet to be sent but the packet send
            was not confirmed
        :raises ChecksumNotImplemented: Right now, only a subset of the checksums specified for
            the CFDP standard are implemented.
        :raises SourceFileDoesNotExist: The source file for which a transaction was requested
            does not exist.
        """
        if self.states.state == CfdpStates.IDLE:
            return FsmResult(self.pdu_holder, self.states)
        elif self.states.state == CfdpStates.BUSY_CLASS_1_NACKED:
            fsm_res = self.__fsm_naked()
            if fsm_res is not None:
                return fsm_res
        return FsmResult(self.pdu_holder, self.states)

    def confirm_packet_sent_advance_fsm(self):
        """Helper method which performs both :py:meth:`confirm_packet_sent` and
        :py:meth:`advance_fsm`
        """
        self.confirm_packet_sent()
        self.advance_fsm()

    def confirm_packet_sent(self):
        """Confirm that a packet generated by the :py:meth:`operation` was sent successfully"""
        self.states.packet_ready = False

    def advance_fsm(self):
        """Advance the internal FSM. This call is necessary to walk through the various steps
        of a CFDP transaction. This step is not done in the main :py:meth:`operation` call
        because the packets generated by this method need to be sent first and then confirmed
        via the :py:meth:`confirm_packet_sent` function.

        :return:
        """
        if self.states.packet_ready:
            raise PacketSendNotConfirmed(
                f"Must send current packet {self.pdu_holder.base} before "
                f"advancing state machine"
            )
        if self.states.state == CfdpStates.BUSY_CLASS_1_NACKED:
            if self.states.step == TransactionStep.SENDING_METADATA:
                self.states.step = TransactionStep.SENDING_FILE_DATA
            elif self.states.step == TransactionStep.SENDING_FILE_DATA:
                self._handle_file_data_sent()
            elif self.states.step == TransactionStep.SENDING_EOF:
                self._handle_eof_sent()

    def reset(self):
        self.states.step = TransactionStep.IDLE
        self.states.state = CfdpStates.IDLE
        self._params.reset()

    def _handle_eof_sent(self):
        if self.cfg.indication_cfg.eof_sent_indication_required:
            self.user.eof_sent_indication(self._params.transaction)
        if self.states.state == CfdpStates.BUSY_CLASS_1_NACKED:
            if self._params.closure_requested:
                if self._params.remote_cfg.check_limit is not None:
                    self._params.check_limit = (
                        self._params.remote_cfg.check_limit.provide_check_limit(
                            local_entity_id=self.cfg.local_entity_id,
                            remote_entity_id=self._params.remote_cfg.entity_id,
                            entity_type=EntityType.SENDING,
                        )
                    )
                self.states.step = TransactionStep.WAIT_FOR_FINISH
            else:
                self.states.step = TransactionStep.NOTICE_OF_COMPLETION
        else:
            self.states.step = TransactionStep.WAIT_FOR_ACK

    def _handle_file_data_sent(self):
        if self._params.fp.progress == self._params.fp.file_size:
            self.states.step = TransactionStep.SENDING_EOF

    def _handle_wait_for_ack(self):
        if self.states.state != CfdpStates.BUSY_CLASS_2_ACKED:
            _LOGGER.error(
                f"Invalid ACK waiting function call for state {self.states.state}"
            )
        pdu_list = self._rec_dict.get(DirectiveType.ACK_PDU)
        if pdu_list is None:
            return FsmResult(self.pdu_holder, self.states)
        for pdu in pdu_list:
            holder = PduHolder(pdu)
            ack_pdu = holder.to_ack_pdu()
            if ack_pdu.directive_code_of_acked_pdu == DirectiveType.EOF_PDU:
                if ack_pdu.condition_code_of_acked_pdu != ConditionCode.NO_ERROR:
                    # TODO: This is required for class 2 transfers. It might make sense
                    #       to remember the condition code of the sent EOF PDU for a basic
                    #       equality check here
                    pass
                self.states.step = TransactionStep.WAIT_FOR_FINISH

    def _notice_of_completion(self):
        if self.cfg.indication_cfg.transaction_finished_indication_required:
            indication_params = TransactionFinishedParams(
                transaction_id=self._params.transaction,
                condition_code=ConditionCode.NO_ERROR,
                file_status=FileDeliveryStatus.FILE_STATUS_UNREPORTED,
                delivery_code=DeliveryCode.DATA_COMPLETE,
            )
            self.user.transaction_finished_indication(indication_params)
        # Transaction finished
        self.reset()

    def _handle_wait_for_finish(self):
        if not self._params.closure_requested:
            _LOGGER.error(
                "Invalid Finish PDU waiting function call, no closure requested"
            )
        # TODO: If transaction closure is requested, a user transaction can only be marked
        #       as finished if TX finished PDU was received. This might take a long time
        #       so it might make sense to think about storing the current state of
        #       the transaction in a source state config file which can be restored
        #       when re-starting the application
        if self._params.closure_requested:
            if self._params.check_limit is not None:
                if self._params.check_limit.timed_out():
                    self._declare_fault(ConditionCode.CHECK_LIMIT_REACHED)
                    _LOGGER.warning(
                        f"Check limit countdown: {self._params.check_limit}"
                    )
            # Check all entries for some robustness against out-of-order reception
            if DirectiveType.FINISHED_PDU in self._rec_dict:
                pdu_list = self._rec_dict.get(DirectiveType.FINISHED_PDU)
                for pdu in pdu_list:
                    holder = PduHolder(pdu)
                    finish_pdu = holder.to_finished_pdu()
                    # TODO: I think there are some more conditions where we can issue a notice
                    #       of completion
                    if finish_pdu.condition_code == ConditionCode.NO_ERROR:
                        self.states.step = TransactionStep.NOTICE_OF_COMPLETION
                    else:
                        # TODO: Implement error handling
                        _LOGGER.warning(
                            f"Received condition code {finish_pdu.condition_code} in "
                            f"Finished PDU"
                        )

    def _setup_transmission_mode(self):
        put_req = self._current_req.to_put_request()
        # Transmission mode settings in the put request override settings from the remote MIB
        if put_req.cfg.trans_mode is not None:
            trans_mode_to_set = put_req.cfg.trans_mode
        else:
            trans_mode_to_set = self._params.remote_cfg.default_transmission_mode
        self._params.transmission_mode = trans_mode_to_set
        if put_req.cfg.closure_requested is not None:
            closure_req_to_set = put_req.cfg.closure_requested
        else:
            closure_req_to_set = self._params.remote_cfg.closure_requested
        self._params.crc_helper.checksum_type = self._params.remote_cfg.crc_type
        self._params.closure_requested = closure_req_to_set

    def _transaction_start(self, put_req: PutRequest):
        if not put_req.cfg.source_file.exists():
            # TODO: Handle this exception in the handler, reset CFDP state machine
            raise SourceFileDoesNotExist(put_req.cfg.source_file)
        size = put_req.cfg.source_file.stat().st_size
        if size == 0:
            self._params.fp.no_file_data = True
        else:
            self._params.fp.file_size = size
        self._params.fp.segment_len = self._params.remote_cfg.max_file_segment_len
        self._params.remote_cfg = self._params.remote_cfg
        self._get_next_transfer_seq_num()
        self._params.transaction = TransactionId(
            source_entity_id=self.cfg.local_entity_id,
            transaction_seq_num=self.transaction_seq_num,
        )
        self.user.transaction_indication(self._params.transaction)

    def _prepare_metadata_pdu(self, put_req: PutRequest):
        if self.states.packet_ready:
            raise PacketSendNotConfirmed(
                f"Must send current packet {self.pdu_holder.base} first"
            )
        self._params.pdu_conf.seg_ctrl = put_req.cfg.seg_ctrl
        self._params.pdu_conf.dest_entity_id = put_req.cfg.destination_id
        self._params.pdu_conf.crc_flag = self._params.remote_cfg.crc_on_transmission
        self._params.pdu_conf.direction = Direction.TOWARDS_RECEIVER
        params = MetadataParams(
            dest_file_name=put_req.cfg.dest_file,
            source_file_name=put_req.cfg.source_file.as_posix(),
            checksum_type=self._params.crc_helper.checksum_type,
            closure_requested=self._params.closure_requested,
            file_size=self._params.fp.file_size,
        )
        self.pdu_holder.base = MetadataPdu(
            pdu_conf=self._params.pdu_conf, params=params
        )

    def _prepare_next_file_data_pdu(self, request: PutRequest) -> bool:
        """Prepare the next file data PDU

        :param request:
        :return: True if a packet was prepared, False if PDU handling is done and the next steps
            in the Copy File procedure can be performed
        """
        # No need to send a file data PDU for an empty file
        if self._params.fp.no_file_data:
            return False
        with open(request.cfg.source_file, "rb") as of:
            if self._params.fp.progress == self._params.fp.file_size:
                return False
            if self.states.packet_ready:
                raise PacketSendNotConfirmed(
                    f"Must send current packet {self.pdu_holder.base} first"
                )
            if self._params.fp.file_size < self._params.fp.segment_len:
                read_len = self._params.fp.file_size
            else:
                if (
                    self._params.fp.progress + self._params.fp.segment_len
                    > self._params.fp.file_size
                ):
                    read_len = self._params.fp.file_size - self._params.fp.progress
                else:
                    read_len = self._params.fp.segment_len
            file_data = self.user.vfs.read_from_opened_file(
                of, self._params.fp.progress, read_len
            )
            # TODO: Support for record continuation state not implemented yet. Segment metadata
            #       flag is therefore always set to False. Segment metadata support also omitted
            #       for now. Implementing those generically could be done in form of a callback,
            #       e.g. abstractmethod of this handler as a first way, another one being
            #       to expect the user to supply some helper class to split up a file
            fd_params = FileDataParams(
                file_data=file_data,
                offset=self._params.fp.progress,
                segment_metadata_flag=False,
            )
            file_data_pdu = FileDataPdu(
                pdu_conf=self._params.pdu_conf, params=fd_params
            )
            self._params.fp.progress += read_len
            self.pdu_holder.base = file_data_pdu
        return True

    def _prepare_eof_pdu(self):
        if self.states.packet_ready:
            raise PacketSendNotConfirmed(
                f"Must send current packet {self.pdu_holder.base} first"
            )
        self.pdu_holder.base = EofPdu(
            file_checksum=self._params.fp.crc32,
            file_size=self._params.fp.file_size,
            pdu_conf=self._params.pdu_conf,
        )

    def _get_next_transfer_seq_num(self):
        next_seq_num = self.seq_num_provider.get_and_increment()
        if self.seq_num_provider.max_bit_width not in [8, 16, 32]:
            raise ValueError(
                "Invalid bit width for sequence number provider, must be one of [8, 16, 32]"
            )
        self._params.pdu_conf.transaction_seq_num = ByteFieldGenerator.from_int(
            self.seq_num_provider.max_bit_width // 8, next_seq_num
        )

    def _declare_fault(self, cond: ConditionCode):
        _LOGGER.warning(
            f"Fault with condition code {cond} was declared for "
            f"transaction {self._params.transaction}"
        )
        fh = self.cfg.default_fault_handlers.get_fault_handler(cond)
        if fh == FaultHandlerCode.NOTICE_OF_CANCELLATION:
            self._notice_of_cancellation()
        elif fh == FaultHandlerCode.NOTICE_OF_SUSPENSION:
            self._notice_of_suspension()
        elif fh == FaultHandlerCode.ABANDON_TRANSACTION:
            self._abandon_transaction()
        self.cfg.default_fault_handlers.report_fault(cond)

    def _notice_of_cancellation(self):
        # TODO: Implement
        pass

    def _notice_of_suspension(self):
        # TODO: Implement
        pass

    def _abandon_transaction(self):
        # TODO: Implement
        pass
